@@ -30,7 +30,7 @@
 #include <X11/Xmu/Converters.h>
 #include <stdio.h>
 #include <ctype.h>
-#include "HandP.h"
+#include <Xkw/HandP.h>
 
 #define offset(field)	XtOffsetOf(HandRec, hand.field)
 
@@ -50,9 +50,9 @@ static XtResource resources[] = {
     {XtNinternalBorderWidth, XtCInternalBorderWidth, XtRDimension, sizeof (Dimension),
      offset(internal_border), XtRImmediate, (XtPointer) 0},
     {XtNnumRows, XtCNumRows, XtRDimension, sizeof (Dimension),
-     offset(num_rows), XtRImmediate, (XtPointer) 0 },
+     offset(num_rows), XtRImmediate, (XtPointer) 1 },
     {XtNnumCols, XtCNumCols, XtRDimension, sizeof (Dimension),
-     offset(num_cols), XtRImmediate, (XtPointer) 0 },
+     offset(num_cols), XtRImmediate, (XtPointer) 1 },
     {XtNrowsHint, XtCRowsHint, XtRBoolean, sizeof (Boolean),
      offset(rows_hint), XtRImmediate, (XtPointer) False},
     {XtNcolsHint, XtCColsHint, XtRBoolean, sizeof (Boolean),
@@ -73,13 +73,30 @@ static XtResource resources[] = {
      offset(col_insert), XtRImmediate, (XtPointer) False},
     {XtNimmediateUpdate, XtCImmediateUpdate, XtRBoolean, sizeof (Boolean),
      offset(immediate_update), XtRImmediate, (XtPointer) True},
+    {XtNselect, XtCSelect, XtRHandSelect, sizeof(HandSelect),
+     offset(select), XtRImmediate, (XtPointer) HandSelectOverlap},
 };
 
 #undef offset
 
-static char defaultTranslations[] =  "<BtnDown>: select()";
+static char defaultTranslations[] =
+    "Shift<BtnDown>: expand()\n"
+    "Shift<BtnUp>: stop()\n"
+    "<Btn4Down>: zoomout()\n"
+    "<Btn4Up>:\n"
+    "<Btn5Down>: zoomin()\n"
+    "<Btn5Up>:\n"
+    "<Btn1Down>: start()\n"
+    "<Btn1Motion>: drag()\n"
+    "<Btn1Up>: stop()\n"
+    "<Btn2Down>: start()\n"
+    "<Btn2Motion>: drag()\n"
+    "<Btn2Up>: stop()\n"
+    "<Btn3Down>: start()\n"
+    "<Btn3Motion>: drag()\n"
+    "<Btn3Up>: stop()";
 
-#define SuperClass  ((SimpleWidgetClass)&simpleClassRec)
+#define SuperClass  ((KSimpleWidgetClass)&ksimpleClassRec)
 
 static int
 handWidth (HandWidget w, int num_cols)
@@ -150,8 +167,8 @@ ColsInRow (HandWidget w, int row)
 {
     int	    maxCol = -1;
     CardPtr c;
-    for (c = w->hand.bottomCard; c; c = c->prev)
-	if (c->shouldBeUp && c->row == row && c->col > maxCol)
+    xkw_foreach_rev(c, &w->hand.cards, list)
+	if (c->row == row && c->col > maxCol)
 	    maxCol = c->col;
     return maxCol;
 }
@@ -161,8 +178,8 @@ RowsInCol (HandWidget w, int col)
 {
     int	maxRow = -1;
     CardPtr c;
-    for (c = w->hand.bottomCard; c; c = c->prev)
-	if (c->shouldBeUp && c->col == col && c->row > maxRow)
+    xkw_foreach_rev(c, &w->hand.cards, list)
+	if (c->col == col && c->row > maxRow)
 	    maxRow = c->row;
     return maxRow;
 }
@@ -187,9 +204,9 @@ XPos (HandWidget w, int row, int col)
 	lastCol = 0;
 	offset = 0;
 	numDefault = 0;
-	for (c = w->hand.bottomCard; c; c = c->prev)
+        xkw_foreach_rev(c, &w->hand.cards, list)
 	{
-	    if (c->shouldBeUp && c->row == row && c->col >= lastCol && c->col < col)
+	    if (c->row == row && c->col >= lastCol && c->col < col)
 	    {
 		numDefault += (c->col - lastCol);
 		if (c->offset == XkwHandDefaultOffset ||
@@ -262,9 +279,9 @@ YPos (HandWidget w, int row, int col)
 	lastRow = 0;
 	offset = 0;
 	numDefault = 0;
-	for (c = w->hand.bottomCard; c; c = c->prev)
+        xkw_foreach_rev(c, &w->hand.cards, list)
 	{
-	    if (c->shouldBeUp && c->col == col && c->row >= lastRow && c->row < row)
+	    if (c->col == col && c->row >= lastRow && c->row < row)
 	    {
 		numDefault += (c->row - lastRow);
 		if (c->offset == XkwHandDefaultOffset ||
@@ -317,6 +334,56 @@ RowFromY (HandWidget w, int y, int col)
     }
 }
 
+static void
+preferredSize(HandWidget w, Dimension *width, Dimension *height)
+{
+    CardPtr c;
+    int max_x = 0, max_y = 0;
+    xkw_foreach_rev(c, &w->hand.cards, list) {
+        int x = XPos(w, c->row, c->col);
+        int y = YPos(w, c->row, c->col);
+        if (x > max_x) max_x = x;
+        if (y > max_y) max_y = y;
+    }
+    *width = max_x + w->hand.card_width;
+    *height = max_y + w->hand.card_height;
+}
+
+
+static Boolean
+CvtStringToHandSelect (Display *dpy,
+                       XrmValue *args,
+                       Cardinal *num_args,
+                       XrmValue *from,
+                       XrmValue *to,
+                       XtPointer *converter_data)
+{
+    char        *s = (char *) from->addr;
+    HandSelect  *result = (HandSelect *) to->addr;
+
+    (void) dpy;
+    (void) args;
+    (void) num_args;
+    (void) converter_data;
+    if (!strcmp (s, "overlap"))
+	*result = HandSelectOverlap;
+    else if (!strcmp (s, "one"))
+	*result = HandSelectOne;
+    else if (!strcmp (s, "all"))
+	*result = HandSelectAll;
+    else
+	return FALSE;
+    return TRUE;
+}
+
+static void
+ClassInitialize(void)
+{
+    XtSetTypeConverter (XtRString, XtRHandSelect, CvtStringToHandSelect,
+                        NULL, (Cardinal)0, XtCacheNone,
+                        (XtDestructor)NULL);
+}
+
 static void
 Initialize (Widget greq, Widget gnew, Arg *args, Cardinal *count)
 {
@@ -328,10 +395,8 @@ Initialize (Widget greq, Widget gnew, Arg *args, Cardinal *count)
     getHandSize (req, &new->core.width, &new->core.height);
     new->hand.real_row_offset = new->hand.row_offset;
     new->hand.real_col_offset = new->hand.col_offset;
-    new->hand.topCard = NULL;
-    new->hand.bottomCard = NULL;
-    new->hand.erased = NULL;
-    new->hand.exposeTime = 0;
+    new->hand.damage = NULL;
+    xkw_list_init(&new->hand.cards);
 }
 
 #define MotionMask ( \
@@ -347,37 +412,22 @@ Initialize (Widget greq, Widget gnew, Arg *args, Cardinal *count)
 static void
 Realize (Widget widget, XtValueMask *value_mask, XSetWindowAttributes *attributes)
 {
-    unsigned int    event_mask = 0;
-#define MAX_BUT	256
-    unsigned char   mapping[MAX_BUT];
-    int	    i, max;
+    *value_mask |= CWBitGravity;
+    attributes->bit_gravity = NorthWestGravity;
 
     (*SuperClass->core_class.realize)(widget, value_mask, attributes);
-    if (*value_mask & CWEventMask)
-	event_mask = attributes->event_mask;
-    event_mask &= PointerGrabMask;
-    if (event_mask & ButtonPressMask)
-    {
-	max = XGetPointerMapping (XtDisplay (widget), mapping, MAX_BUT);
-	for (i = 0; i < max; i++)
-	{
-	    if (mapping[i] != 0)
-		XtGrabButton (widget, i, AnyModifier, True, event_mask,
-			      GrabModeAsync, GrabModeAsync, None, None);
-	}
-    }
 }
 
 static void
 Destroy (Widget gw)
 {
     HandWidget	w = (HandWidget) gw;
-    CardPtr	c, n;
+    CardPtr	c;
 
-    for (c = w->hand.topCard; c; c = n) {
-	n = c->next;
+    xkw_foreach(c, &w->hand.cards, list)
 	Dispose (c);
-    }
+    if (w->hand.update_proc)
+        XtRemoveWorkProc(w->hand.update_proc);
 }
 
 static XtGeometryResult
@@ -425,322 +475,215 @@ static CardPtr XYToCard (HandWidget w, int x, int y)
 {
     CardPtr c;
 
-    for (c = w->hand.topCard; c; c = c->next) {
+    xkw_foreach(c, &w->hand.cards, list) {
 	if (XYInCard (w, c, x, y))
 	    return c;
     }
     return NULL;
 }
 
-static void ActionSelect (Widget gw, XEvent *event, String *params, Cardinal *num_params)
+/* Start location information */
+static HandLocation    start_location;     /* where the drag started */
+static Boolean         expanding;
+
+static void
+DoInputCallback(Widget gw, Action action,
+		XEvent *event, String *params, Cardinal *num_params)
 {
     HandWidget	    w = (HandWidget) gw;
     CardPtr	    c;
     HandInputRec    input;
     int		    row, col;
     XtPointer	    private;
+    HandAction      hand_action;
+    Position        x = event->xmotion.x;
+    Position        y = event->xmotion.y;
 
-    if (event->type == KeyPress && *num_params > 1)
+    c = XYToCard (w, x, y);
+    if (c)
+    {
+	row = c->row;
+	col = c->col;
+	private = c->private;
+    }
+    else
     {
 	if (w->hand.row_major)
 	{
-	    row = atoi (params[1]);
-	    col = w->hand.num_cols - 1;
+	    row = RowFromY (w, y, -1);
+	    col = ColFromX (w, x, row);
 	}
 	else
 	{
-	    col = atoi (params[1]);
-	    row = w->hand.num_rows - 1;
+	    col = ColFromX (w, x, -1);
+	    row = RowFromY (w, y, col);
 	}
+	if (row < 0)
+	    row = 0;
+	else if (w->hand.row_major && w->hand.num_rows <= row)
+	    row = w->hand.num_rows - 1;
+	if (col < 0)
+	    col = 0;
+	else if (!w->hand.row_major && w->hand.num_cols <= col)
+	    col = w->hand.num_cols - 1;
 	private = 0;
     }
-    else
-    {
-	c = XYToCard (w, event->xbutton.x, event->xbutton.y);
-	if (c)
-	{
-	    row = c->row;
-	    col = c->col;
-	    private = c->private;
-	}
-	else
-	{
-	    if (w->hand.row_major)
-	    {
-		row = RowFromY (w, event->xbutton.y, -1);
-		col = ColFromX (w, event->xbutton.x, row);
-	    }
-	    else
-	    {
-		col = ColFromX (w, event->xbutton.x, -1);
-		row = RowFromY (w, event->xbutton.y, col);
-	    }
-	    if (row < 0)
-		row = 0;
-	    else if (w->hand.row_major && w->hand.num_rows <= row)
-		row = w->hand.num_rows - 1;
-	    if (col < 0)
-		col = 0;
-	    else if (!w->hand.row_major && w->hand.num_cols <= col)
-		col = w->hand.num_cols - 1;
-	    private = 0;
-	}
+    input.current.w = gw;
+    input.current.x = x;
+    input.current.y = y;
+    input.current.row = row;
+    input.current.col = col;
+    input.current.card = c;
+    input.current.private = private;
+    input.current.dragging = FALSE;
+    switch (action) {
+    case ActionExpand:
+        start_location = input.current;
+        hand_action = HandActionExpand;
+        expanding = TRUE;
+        break;
+    case ActionStart:
+        start_location = input.current;
+        hand_action = HandActionStart;
+        expanding = FALSE;
+        break;
+    case ActionDrag: {
+        Position        w_x = x;
+        Position        w_y = y;
+        Position        dx, dy;
+        Dimension       motion_min = (w->hand.card_width * w->hand.card_height) / 100;
+
+        XkwTranslateCoordsPosition(start_location.w, gw, &w_x, &w_y);
+        dx = w_x - start_location.x;
+        dy = w_y - start_location.y;
+        hand_action = HandActionStart;
+
+        if (dx * dx + dy * dy >= motion_min)
+            start_location.dragging = TRUE;
+        if (!start_location.dragging)
+            return;
+        break;
     }
-    input.w = gw;
-    input.row = row;
-    input.col = col;
-    input.private = private;
+    case ActionStop:
+        if (expanding)
+            hand_action = HandActionUnexpand;
+        else if (start_location.dragging)
+            hand_action = HandActionDrag;
+        else
+            hand_action = HandActionClick;
+        break;
+    }
+
+    input.start = start_location;
+    input.action = hand_action;
     input.params = params;
-    input.event = *event;
     input.num_params = num_params;
-    XtCallCallbackList ((Widget) w, w->hand.input_callback, (XtPointer) &input);
+
+    HandDrag(&input, action);
+    if (action != ActionDrag)
+        XtCallCallbackList ((Widget) w, w->hand.input_callback, (XtPointer) &input);
 }
 
-static Bool CardInRect (HandWidget w, XRectangle *rect, CardPtr c)
+static void StartAction (Widget gw, XEvent *event, String *params, Cardinal *num_params)
 {
-    XRectangle	*card = &c->clip;
-
-    (void) w;
-    return rect->x < card->x + card->width &&
-	   card->x < rect->x + rect->width &&
-	   rect->y < card->y + card->height &&
-	   card->y < rect->y + rect->height;
+    DoInputCallback(gw, ActionStart, event, params, num_params);
 }
 
-#define ClipRects(allClipped) { \
-    if (cx1 <= x1 && x2 <= cx2) \
-    { \
-	if (cy1 <= y1 && y1 < cy2) \
-	    y1 = cy2; \
-	if (cy1 < y2 && y2 <= cy2) \
-	    y2 = cy1; \
-	if (y2 < y1) \
-	    y2 = y1; \
-	if (y1 == y2) \
-	    allClipped; \
-    } \
-    if (cy1 <= y1 && y2 <= cy2) \
-    { \
-	if (cx1 <= x1 && x1 < cx2) \
-	    x1 = cx2; \
-	if (cx1 < x2 && x2 <= cx2) \
-	    x2 = cx1; \
-	if (x2 < x1) \
-	    x2 = x1; \
-	if (x1 == x2) \
-	    allClipped; \
-    } \
-}
-
-static void MarkCard (HandWidget w, CardPtr c)
+static void DragAction (Widget gw, XEvent *event, String *params, Cardinal *num_params)
 {
-    XRectangle	*r;
-    ErasedPtr	e, *p;
-    int		cx1, cy1, cx2, cy2;
-    int		x1, y1, x2, y2;
-    Boolean	allClipped;
-
-    c->redisplay = TRUE;
-    if (w->hand.erased)
-    {
-	cx1 = c->x + w->hand.display_x;
-	cy1 = c->y + w->hand.display_y;
-	cx2 = cx1 + w->hand.display_width;
-	cy2 = cy1 + w->hand.display_height;
-	if (cx1 < c->clip.x)
-	    cx1 = c->clip.x;
-	if (cx2 > c->clip.x + c->clip.width)
-	    cx2 = c->clip.x + c->clip.width;
-	if (cy1 < c->clip.y)
-	    cy1 = c->clip.y;
-	if (cy2 > c->clip.y + c->clip.height)
-	    cy2 = c->clip.y + c->clip.height;
-	for (p = &w->hand.erased; (e = *p);)
-	{
-	    allClipped = False;
-	    x1 = e->r.x;
-	    x2 = x1 + e->r.width;
-	    y1 = e->r.y;
-	    y2 = y1 + e->r.height;
-	    ClipRects (allClipped = True);
-	    if (allClipped)
-	    {
-		*p = e->next;
-		Dispose (e);
-	    }
-	    else
-	    {
-		e->r.x = x1;
-		e->r.y = y1;
-		e->r.width = x2 - x1;
-		e->r.height = y2 - y1;
-		p = &e->next;
-	    }
-	}
-    }
-    r = &c->clip;
-    for (c = c->prev; c; c = c->prev)
-	if (!c->redisplay && CardInRect (w, r, c))
-	    MarkCard (w, c);
+    DoInputCallback(gw, ActionDrag, event, params, num_params);
 }
 
-static void MarkRectangle (HandWidget w, XRectangle *r)
+static void StopAction (Widget gw, XEvent *event, String *params, Cardinal *num_params)
 {
-    CardPtr c;
-
-    for (c = w->hand.bottomCard; c; c = c->prev)
-	if (!c->redisplay && CardInRect (w, r, c))
-	    MarkCard (w, c);
+    if (!XkwForwardEvent(NULL, gw, event))
+	DoInputCallback(gw, ActionStop, event, params, num_params);
 }
 
-static void
-EraseCard (HandWidget w, CardPtr c)
+static void ZoomInAction (Widget gw, XEvent *event, String *params, Cardinal *num_params)
 {
-    ErasedPtr	e = New(ErasedRec);
-
-    e->r = c->clip;
-    e->next = w->hand.erased;
-    e->fill = True;
-    e->isCard = True;
-    e->cardX = c->x;
-    e->cardY = c->y;
-    w->hand.erased = e;
+    (void) gw;
+    (void) event;
+    (void) params;
+    (void) num_params;
 }
 
-static void
-ComputeClip (HandWidget w, CardPtr card)
+static void ZoomOutAction (Widget gw, XEvent *event, String *params, Cardinal *num_params)
 {
-    int	    x1, y1, x2, y2;
-    int	    cx1, cy1, cx2, cy2;
-    int	    ox1, oy1, ox2, oy2;
-    CardPtr c;
+    (void) gw;
+    (void) event;
+    (void) params;
+    (void) num_params;
+}
 
-    ox1 = x1 = card->x;
-    oy1 = y1 = card->y;
-    ox2 = x2 = x1 + w->hand.card_width;
-    oy2 = y2 = y1 + w->hand.card_height;
-    c = card;
-    while ((c = c->prev))
-    {
-	if (!c->shouldBeUp)
-	    continue;
-	cx1 = c->x + w->hand.display_x;
-	cy1 = c->y + w->hand.display_y;
-	cx2 = cx1 + w->hand.display_width;
-	cy2 = cy1 + w->hand.display_height;
-	ClipRects (break);
-    }
-    card->clip.x = x1;
-    card->clip.y = y1;
-    card->clip.width = x2 - x1;
-    card->clip.height = y2 - y1;
-    if (x1 == ox1 && y1 == oy1 && x2 == ox2 && y2 == oy2)
-	card->clipped = ClipUnclipped;
-    else if (x2 > x1 && y2 > y1)
-	card->clipped = ClipPartclipped;
-    else
-	card->clipped = ClipAllclipped;
+static void ExpandAction (Widget gw, XEvent *event, String *params, Cardinal *num_params)
+{
+    DoInputCallback(gw, ActionExpand, event, params, num_params);
 }
 
 void
-HandUpdateDisplay (Widget gw)
+HandCardRectangle(HandWidget w, CardPtr card, XRectangle *c)
 {
-    HandWidget	    w = (HandWidget) gw;
-    CardPtr	    c, nc;
-    ErasedPtr	    e;
-    HandDisplayRec  display;
-    int		    x, y;
+    c->x = card->x;
+    c->y = card->y;
+    c->width = w->hand.card_width;
+    c->height = w->hand.card_height;
+}
 
-    if (!XtIsRealized (gw))
-	return;
+static void
+DamageCard(HandWidget w, CardPtr card)
+{
+    XRectangle c;
 
-    /* Mark cards for redisplay */
-    for (c = w->hand.topCard; c; c = nc)
-    {
-	nc = c->next;
-	if (c->shouldBeUp)
-	{
-	    x = XPos (w, c->row, c->col);
-	    y = YPos (w, c->row, c->col);
-	    if (c->isUp)
-	    {
-		if (c->x != x || c->y != y)
-		{
-		    EraseCard (w, c);
-		    c->isUp = False;
-		}
-	    }
+    if (card->shown) {
+	HandCardRectangle(w, card, &c);
+	if (!w->hand.damage)
+	    w->hand.damage = XCreateRegion();
+	XUnionRectWithRegion(&c, w->hand.damage, w->hand.damage);
+    }
+}
+
+Bool
+HandCardInRegion (HandWidget w, CardPtr card, Region region)
+{
+    if (region == NULL)
+	return True;
+
+    XRectangle c;
+    HandCardRectangle(w, card, &c);
+    return XRectInRegion(region, c.x, c.y, c.width, c.height);
+}
+
+static void
+UpdateCards (HandWidget w)
+{
+    CardPtr c;
+    xkw_foreach_rev(c, &w->hand.cards, list) {
+	int x = XPos(w, c->row, c->col);
+	int y = YPos(w, c->row, c->col);
+
+	if (!c->shown || x != c->x || y != c->y) {
+	    if (c->shown)
+		DamageCard(w, c);
 	    c->x = x;
 	    c->y = y;
-	    ComputeClip (w, c);
-	}
-	else if (c->isUp)
-	    EraseCard (w, c);
-	if (c->delete)
-	{
-	    if (c->prev)
-		c->prev->next = c->next;
-	    else
-		w->hand.topCard = c->next;
-	    if (c->next)
-		c->next->prev = c->prev;
-	    else
-		w->hand.bottomCard = c->prev;
-	    Dispose (c);
-	}
-    }
-    /* mark effects of redisplay */
-    for (c = w->hand.bottomCard; c; c = c->prev)
-    {
-	if ((c->shouldBeUp && !c->isUp) || c->forceRedraw)
-	{
-	    MarkCard (w, c);
-	    c->isUp = True;
-	    c->forceRedraw = False;
+	    c->shown = True;
+	    DamageCard(w, c);
 	}
     }
+}
 
-    /* clear out the blank areas */
-    while ((e = w->hand.erased))
-    {
-	w->hand.erased = e->next;
-	c = 0;
-	if (e->isCard)
-	{
-	    for (c = w->hand.bottomCard; c; c = c->prev)
-		if (c->redisplay && c->x == e->cardX && c->y == e->cardY)
-		    break;
-	}
-	if (!c)
-	{
-	    MarkRectangle (w, &e->r);
-	    if (e->fill)
-	    {
-		XClearArea (XtDisplay(w), XtWindow (w),
-			    e->r.x, e->r.y, e->r.width, e->r.height, False);
-	    }
-	}
-	Dispose (e);
-    }
+static void
+Redisplay (Widget gw, XEvent *event, Region region);
 
-    /* redisplay cards */
-    for (c = w->hand.bottomCard; c; c = c->prev)
-	if (c->redisplay)
-	{
-	    if (c->clipped != ClipAllclipped)
-	    {
-		display.w = (Widget) w;
-		display.x = c->x;
-		display.y = c->y;
-		display.private = c->private;
-		display.clipped = False;
-		display.clip = c->clip;
-		if (c->clipped == ClipPartclipped)
-		    display.clipped = True;
-		XtCallCallbackList ((Widget) w, w->hand.display_callback, (XtPointer) &display);
-	    }
-	    c->redisplay = FALSE;
-	}
+static void
+EraseTimer(XtPointer client_data, XtIntervalId *timer)
+{
+    Widget gw = (Widget) client_data;
+    HandWidget w = (HandWidget) gw;
+    (void) timer;
+    w->hand.erase_proc = 0;
+    Redisplay(gw, NULL, NULL);
 }
 
 /*
@@ -752,11 +695,75 @@ HandUpdateDisplay (Widget gw)
 static void
 Resize (Widget gw)
 {
-    HandWidget	w = (HandWidget) gw;
+    HandWidget w = (HandWidget) gw;
+    Dimension col_offset, row_offset;
+    Bool erase = FALSE;
 
-    w->hand.real_col_offset = BestColOffset (w, w->hand.num_cols);
-    w->hand.real_row_offset = BestRowOffset (w, w->hand.num_rows);
-    w->hand.exposeTime = 0;
+    if (*SuperClass->core_class.resize != NULL)
+	(*SuperClass->core_class.resize)(gw);
+
+    /* Reschedule the erase if we resize again */
+    if (w->hand.erase_proc)
+        erase = TRUE;
+
+    col_offset = BestColOffset (w, w->hand.num_cols);
+    row_offset = BestRowOffset (w, w->hand.num_rows);
+
+    /* Check to see if the global offsets change */
+    if (col_offset != w->hand.real_col_offset ||
+        row_offset != w->hand.real_row_offset)
+    {
+        erase = TRUE;
+        w->hand.real_col_offset = col_offset;
+        w->hand.real_row_offset = row_offset;
+    }
+
+    /* Check to see if any card will move */
+    if (!erase && (w->hand.rows_hint || w->hand.cols_hint)) {
+        CardPtr c;
+        xkw_foreach(c, &w->hand.cards, list) {
+            int x = XPos(w, c->row, c->col);
+            int y = YPos(w, c->row, c->col);
+
+            if (x != c->x || y != c->y)
+                erase = TRUE;
+        }
+    }
+
+    /* Schedule a repaint */
+    if (XtIsRealized(gw) && erase) {
+        if (w->hand.erase_proc)
+            XtRemoveTimeOut(w->hand.erase_proc);
+        w->hand.erase_proc = XtAddTimeOut(10, EraseTimer, gw);
+    }
+}
+
+static void
+Paint (HandWidget w, Region region)
+{
+    if (XtIsRealized((Widget) w)) {
+	cairo_t 	*cr = XkwDrawBegin ((Widget) w, region);
+	HandDisplayRec	display;
+	CardPtr		c;
+
+	display.w = (Widget) w;
+	display.cr = cr;
+	/* redisplay cards */
+        xkw_foreach_rev(c, &w->hand.cards, list) {
+	    if (!c->hidden && HandCardInRegion(w, c, region)) {
+		cairo_save(cr);
+		cairo_translate(cr, c->x, c->y);
+		display.private = c->private;
+		XtCallCallbackList ((Widget) w, w->hand.display_callback, (XtPointer) &display);
+		cairo_restore(cr);
+	    }
+	}
+	XkwDrawEnd((Widget) w, region, cr);
+    }
+    if (w->hand.damage) {
+	XDestroyRegion(w->hand.damage);
+	w->hand.damage = NULL;
+    }
 }
 
 /*
@@ -767,40 +774,20 @@ static void
 Redisplay (Widget gw, XEvent *event, Region region)
 {
     HandWidget	    w = (HandWidget) gw;
-    ErasedPtr	    e;
-    int		    cx1, cy1, cx2, cy2;
-    int		    x1, y1, x2, y2;
 
-    (void) region;
-    x1 = event->xexpose.x;
-    x2 = x1 + event->xexpose.width;
-    y1 = event->xexpose.y;
-    y2 = y1 + event->xexpose.height;
-    if (event->xexpose.serial <= w->hand.exposeTime)
-    {
-	cx1 = w->hand.lastExpose.x;
-	cy1 = w->hand.lastExpose.y;
-	cx2 = cx1 + w->hand.lastExpose.width;
-	cy2 = cy1 + w->hand.lastExpose.height;
-	ClipRects (return)
-    }
-    e = New(ErasedRec);
-    e->r.x = x1;
-    e->r.y = y1;
-    e->r.width = x2 - x1;
-    e->r.height = y2 - y1;
-    e->next = w->hand.erased;
-    e->fill = False;
-    e->isCard = False;
-    w->hand.erased = e;
-    if (event->xexpose.count == 0) {
-	w->hand.exposeTime = NextRequest (XtDisplay (w));
-	w->hand.lastExpose.x = x1;
-	w->hand.lastExpose.y = y1;
-	w->hand.lastExpose.width = x2 - x1;
-	w->hand.lastExpose.height = y2 - y1;
-	HandUpdateDisplay (gw);
+    (void) event;
+    if (!XtIsRealized (gw))
+	return;
+
+    if (w->hand.erase_proc)
+        return;
+
+    UpdateCards(w);
+    if (region && w->hand.damage) {
+	XUnionRegion(region, w->hand.damage, w->hand.damage);
+	region = w->hand.damage;
     }
+    Paint (w, region);
 }
 
 /*
@@ -816,21 +803,53 @@ SetValues (Widget gcur, Widget greq, Widget gnew, Arg *args, Cardinal *count)
  		new = (HandWidget) gnew;
     Dimension	curwidth, curheight, reqwidth, reqheight;
     Dimension	width, height;
+    Bool force_resize = FALSE;
 
     (void) args;
     (void) count;
+    (void) new;
+    if (cur->hand.num_rows != req->hand.num_rows ||
+        cur->hand.num_cols != req->hand.num_cols ||
+        cur->hand.row_offset != req->hand.row_offset ||
+        cur->hand.col_offset != req->hand.col_offset)
+        force_resize = TRUE;
     getHandSize (cur, &curwidth, &curheight);
     getHandSize (req, &reqwidth, &reqheight);
-    if (curwidth != reqwidth || curheight != reqheight)
+    if (curwidth != reqwidth || curheight != reqheight || force_resize)
     {
 	XtMakeResizeRequest (gnew, reqwidth, reqheight, &width, &height);
-	if (width != curwidth || height != curheight)
+	if (width != curwidth || height != curheight || force_resize)
 	    Resize (gnew);
     }
-    new->hand.exposeTime = 0;
     return TRUE;
 }
 
+void
+HandUpdateDisplay (Widget gw)
+{
+    HandWidget w = (HandWidget) gw;
+
+    UpdateCards (w);
+    if (w->hand.damage)
+	Paint(w, w->hand.damage);
+}
+
+void
+HandSetPreferredSize (Widget gw)
+{
+    HandWidget w = (HandWidget) gw;
+    Dimension width, height;
+    Dimension curwidth = XtWidth(gw);
+    Dimension curheight = XtHeight(gw);
+
+    preferredSize(w, &width, &height);
+    if (width != curwidth || height != curheight) {
+	XtMakeResizeRequest (gw, width, height, &width, &height);
+	if (width != curwidth || height != curheight)
+	    Resize (gw);
+    }
+}
+
 /* Insert a card */
 XtPointer
 HandAddCard (Widget	gw,
@@ -847,8 +866,8 @@ HandAddCard (Widget	gw,
     if (row == InsertRow)
     {
 	maxPos = -1;
-	for (c = w->hand.topCard; c; c = c->next)
-	{
+        xkw_foreach(c, &w->hand.cards, list)
+        {
 	    if (w->hand.row_major)
 	    {
 		if (c->row > maxPos)
@@ -865,7 +884,7 @@ HandAddCard (Widget	gw,
     if (col == InsertCol)
     {
 	maxPos = -1;
-	for (c = w->hand.topCard; c; c = c->next)
+        xkw_foreach(c, &w->hand.cards, list)
 	{
 	    if (!w->hand.row_major)
 	    {
@@ -881,8 +900,8 @@ HandAddCard (Widget	gw,
 	col = maxPos + 1;
     }
     /* find the card below this one */
-    sib = NULL;
-    for (c = w->hand.bottomCard; c; c = c->prev)
+    sib = (CardPtr) &w->hand.cards;
+    xkw_foreach_rev(c, &w->hand.cards, list)
     {
 	if (w->hand.row_major)
 	{
@@ -900,48 +919,29 @@ HandAddCard (Widget	gw,
 	}
     }
     c = New(CardRec);
-    c->redisplay = False;
-    c->forceRedraw = False;
-    c->isUp = False;
-    c->shouldBeUp = True;
-    c->delete = False;
     c->private = private;
     c->row = row;
     c->col = col;
     c->offset = offset;
+    c->shown = False;
+    c->hidden = False;
     /* insert the new card on the list */
-    c->next = sib;	/* c is above sib */
-
-    if (sib)
-	c->prev = sib->prev;
-    else
-	c->prev = w->hand.bottomCard;
-
-    if (c->next)
-	c->next->prev = c;
-    else
-	w->hand.bottomCard = c;
-
-    if (c->prev)
-	c->prev->next = c;
-    else
-	w->hand.topCard = c;
+    xkw_list_append(&c->list, &sib->list);
 
     /* adjust higher cards rows */
     if (w->hand.row_insert)
     {
-	for (sib = c->prev; sib; sib = sib->prev)
+        xkw_foreach_startat_rev(sib, xkw_prev(c, list), &w->hand.cards, list)
 	    if (sib->col == c->col && sib->row == row)
 		sib->row = ++row;
     }
     /* adjust higher cards columns */
     if (w->hand.col_insert)
     {
-	for (sib = c->prev; sib; sib = sib->prev)
+        xkw_foreach_startat_rev(sib, xkw_prev(c, list), &w->hand.cards, list)
 	    if (sib->row == c->row && sib->col == col)
 		sib->col = ++col;
     }
-    w->hand.exposeTime = 0;
     if (XtIsRealized(gw) && w->hand.immediate_update)
 	HandUpdateDisplay (gw);
     return (XtPointer) c;
@@ -954,29 +954,32 @@ HandRemoveCard (Widget gw, XtPointer card)
     HandWidget	    w = (HandWidget) gw;
     CardPtr	    c, sib;
     int		    row, col;
+    Bool            found = FALSE;
 
-    for (c = w->hand.topCard; c; c = c->next)
-	if (c == (CardPtr) card)
+    xkw_foreach(c, &w->hand.cards, list)
+	if (c == (CardPtr) card) {
+            found = TRUE;
 	    break;
-    if (!c)
+        }
+    if (!found)
 	return;
-    c->shouldBeUp = False;
-    c->delete = True;
     if (w->hand.row_insert)
     {
 	row = c->row;
-	for (sib = c->prev; sib; sib = sib->prev)
+        xkw_foreach_startat_rev(sib, xkw_prev(c, list), &w->hand.cards, list)
 	    if (sib->col == c->col && sib->row == row + 1)
 		sib->row = row++;
     }
     if (w->hand.col_insert)
     {
 	col = c->col;
-	for (sib = c->prev; sib; sib = sib->prev)
+        xkw_foreach_startat_rev(sib, xkw_prev(c, list), &w->hand.cards, list)
 	    if (sib->row == c->row && sib->col == col + 1)
 		sib->col = col++;
     }
-    w->hand.exposeTime = 0;
+    xkw_list_del(&c->list);
+    DamageCard(w, c);
+    Dispose (c);
     if (XtIsRealized (gw) && w->hand.immediate_update)
 	HandUpdateDisplay (gw);
 }
@@ -987,15 +990,18 @@ HandReplaceCard (Widget gw, XtPointer card, XtPointer private, int offset)
 {
     HandWidget	    w = (HandWidget) gw;
     CardPtr c;
-    for (c = w->hand.topCard; c; c = c->next)
-	if (c == (CardPtr) card)
+    Bool found = FALSE;
+
+    xkw_foreach(c, &w->hand.cards, list)
+	if (c == (CardPtr) card) {
+            found = TRUE;
 	    break;
-    if (!c)
+        }
+    if (!found)
 	return;
     c->private = private;
-    c->forceRedraw = True;
     c->offset = offset;
-    w->hand.exposeTime = 0;
+    DamageCard(w, c);
     if (XtIsRealized (gw) && w->hand.immediate_update)
     	HandUpdateDisplay (gw);
 }
@@ -1017,11 +1023,14 @@ HandRectangleForCard (Widget gw, XtPointer card, XRectangle *r)
 {
     HandWidget	    w = (HandWidget) gw;
     CardPtr	c;
+    Bool found = FALSE;
 
-    for (c = w->hand.topCard; c; c = c->next)
-	if (c == (CardPtr) card)
+    xkw_foreach(c, &w->hand.cards, list)
+	if (c == (CardPtr) card) {
+            found = TRUE;
 	    break;
-    if (!c)
+        }
+    if (!found)
     {
 	r->x = 0;
 	r->y = 0;
@@ -1053,21 +1062,79 @@ void
 HandRemoveAllCards (Widget gw)
 {
     HandWidget	    w = (HandWidget) gw;
-    CardPtr	    c, n;
+    CardPtr	    c;
 
-    for (c = w->hand.topCard; c; c = n) {
-	n = c->next;
+    xkw_foreach(c, &w->hand.cards, list)
 	free ((char *) c);
-    }
-    w->hand.topCard = 0;
-    w->hand.bottomCard = 0;
-    w->hand.exposeTime = 0;
+    xkw_list_init(&w->hand.cards);
     if (XtIsRealized (gw))
-	XClearWindow (XtDisplay (w), XtWindow (w));
+	Redisplay(gw, NULL, NULL);
+}
+
+static Boolean
+HandUpdateAllCards(XtPointer p)
+{
+    Widget      gw = (Widget) p;
+    HandWidget  w = (HandWidget) gw;
+
+    w->hand.update_proc = 0;
+    if (XtIsRealized(gw))
+        HandUpdateDisplay(gw);
+    return TRUE;
+}
+
+void
+HandShowAllCards (Widget gw)
+{
+    HandWidget	    w = (HandWidget) gw;
+    CardPtr	    c;
+    Boolean         changed = FALSE;
+
+    xkw_foreach(c, &w->hand.cards, list)
+        if (c->hidden) {
+            c->hidden = FALSE;
+            changed = TRUE;
+            DamageCard(w, c);
+        }
+    if (changed && XtIsRealized (gw)) {
+        if (!w->hand.update_proc)
+            w->hand.update_proc = XtAddWorkProc(HandUpdateAllCards, (XtPointer) gw);
+    }
+}
+
+void
+HandHideCard (Widget gw, CardPtr c)
+{
+    HandWidget	    w = (HandWidget) gw;
+    c->hidden = TRUE;
+    DamageCard(w, c);
+    if (XtIsRealized (gw) && w->hand.immediate_update)
+    	HandUpdateDisplay (gw);
+}
+
+static Boolean
+HandDefaultCardIsEmpty (Widget gw, XtPointer private)
+{
+    (void) gw;
+    (void) private;
+    return FALSE;
+}
+
+Boolean
+HandCardIsEmpty (Widget gw, XtPointer private)
+{
+    HandClassRec *class = (HandClassRec *) (gw->core.widget_class);
+
+    return (*class->hand_class.card_is_empty)(gw, private);
 }
 
 static XtActionsRec actions[] = {
-    { "select", ActionSelect },		    /* select card */
+    { "start", StartAction },		    /* select card */
+    { "drag", DragAction },
+    { "stop", StopAction },
+    { "zoomin", ZoomInAction },
+    { "zoomout", ZoomOutAction },
+    { "expand", ExpandAction },
 };
 
 HandClassRec handClassRec = {
@@ -1075,7 +1142,7 @@ HandClassRec handClassRec = {
     /* superclass		*/	(WidgetClass) SuperClass,
     /* class_name		*/	"Hand",
     /* widget_size		*/	sizeof(HandRec),
-    /* class_initialize		*/	NULL,
+    /* class_initialize		*/	ClassInitialize,
     /* class_part_initialize	*/	NULL,
     /* class_inited		*/	FALSE,
     /* initialize		*/	Initialize,
@@ -1087,7 +1154,7 @@ HandClassRec handClassRec = {
     /* num_resources		*/	XtNumber(resources),
     /* xrm_class		*/	NULLQUARK,
     /* compress_motion		*/	TRUE,
-    /* compress_exposure	*/	FALSE,
+    /* compress_exposure	*/	TRUE,
     /* compress_enterleave	*/	TRUE,
     /* visible_interest		*/	FALSE,
     /* destroy			*/	Destroy,
@@ -1105,12 +1172,19 @@ HandClassRec handClassRec = {
     /* display_accelerator	*/	XtInheritDisplayAccelerator,
     /* extension		*/	NULL
   },
-  { /* simple fields */
-    /* change_sensitive		*/	XtInheritChangeSensitive,
-    /* extension                */      NULL,
+  /* simple */
+  {
+    XtInheritChangeSensitive,		/* change_sensitive */
+#ifndef OLDXAW
+    NULL,                               /* extension */
+#endif
+  },
+  {
+    /* ksimple fields */
+    /* empty			*/	0
   },
   { /* hand fields */
-    /* ignore                   */	0
+      .card_is_empty = HandDefaultCardIsEmpty,
   }
 };
 
