@@ -1100,7 +1100,7 @@ main (int argc, char **argv)
 
     message = XtCreateManagedWidget("message", klabelWidgetClass, frame, NULL, 0);
 
-    srandom(getpid() ^ time((long *) 0));
+    srandom(getpid() ^ time ((time_t *) 0));
     NewGame();
     XtRealizeWidget(toplevel);
     wm_delete_window = XInternAtom(XtDisplay(toplevel), "WM_DELETE_WINDOW",
