@@ -909,7 +909,7 @@ main (int argc, char **argv)
     tail = XtCreateManagedWidget ("tail", cardsWidgetClass, frame, NULL, 0);
     XtAddCallback (tail, XtNinputCallback, InputCallback, NULL);
     message = XtCreateManagedWidget ("message", klabelWidgetClass, frame, NULL, 0);
-    srandom (getpid () ^ time ((long *) 0));
+    srandom (getpid () ^ time ((time_t *) 0));
     NewGame ();
     XtRealizeWidget (toplevel);
     wm_delete_window = XInternAtom(XtDisplay(toplevel), "WM_DELETE_WINDOW",
