@@ -13,25 +13,26 @@
 # include	<X11/Xaw/Paned.h>
 # include	<X11/Xaw/Form.h>
 # include	<X11/Xaw/Box.h>
-# include	<X11/Xaw/Command.h>
+# include	<Xkw/KCommand.h>
 # include	<X11/Xaw/Dialog.h>
-# include	<X11/Xaw/Label.h>
-# include	<X11/Xaw/MenuButton.h>
-# include	<X11/Xaw/SimpleMenu.h>
-# include	<X11/Xaw/SmeBSB.h>
+# include	<Xkw/KLabel.h>
+# include	<Xkw/KMenuButton.h>
+# include	<Xkw/KSimpleMenu.h>
+# include	<Xkw/KSmeBSB.h>
 # include	<X11/Xaw/AsciiText.h>
 # include	<X11/Xaw/Cardinals.h>
 # include	<Xkw/Hand.h>
 # include	<Xkw/Thermo.h>
 # include	<Xkw/Layout.h>
 # include	<Xkw/Pad.h>
+# include	<Xkw/KTextLine.h>
+# include	<Xkw/Xkw.h>
+# include	<Xkw/Message.h>
 # include	<X11/Xutil.h>
 # include	"gray.bm"
-
-#ifdef CTRL
-# undef CTRL
-#endif
-#define CTRL(x)	(x - 'A' + 1)
+# include	"cards-svg.h"
+# include	"Mille-res.h"
+# include       "MilleCards.h"
 
 # include	"card.h"
 
@@ -69,6 +70,7 @@ Widget		human_hand;
 Widget		human_play;
 Widget		human_safeties;
 Widget		human_safety_label;
+double		scale = 2.0;
 
 Widget		yes_or_no_shell;
 Widget		yes_or_no_dialog;
@@ -122,28 +124,16 @@ static struct menuEntry fileMenuEntries[] = {
 
 int	iscolor;
 
-static void
-displayString (Widget w, char *string)
-{
-    Arg   arg[1];
-
-    XtSetArg (arg[0], XtNlabel, string);
-    XtSetValues (w, arg, 1);
-}
-
 void
-Message (char *string)
+MilleMessage (char *string)
 {
-    displayString (message, string);
+    Message (message, "%s", string);
 }
 
 void
 VError(const char *string, va_list ap)
 {
-    char	buf[512];
-
-    vsprintf (buf, string, ap);
-    displayString (errors, buf);
+    MessageV(errors, string, ap);
 }
 
 void
@@ -165,25 +155,29 @@ Prompt (char *string)
 #endif
 }
 
+/*
 void
 debug (int pos, char *string, int a0, int a1, int a2)
 {
 }
+*/
 
 static int  yn_done, yn_answer;
 
-static void YesFunc (w, closure, data)
-    Widget	w;
-    XtPointer	closure, data;
+static void YesFunc (Widget w, XtPointer closure, XtPointer data)
 {
+    (void) w;
+    (void) closure;
+    (void) data;
     yn_answer = 1;
     yn_done = 1;
 }
 
-static void NoFunc (w, closure, data)
-    Widget	w;
-    XtPointer	closure, data;
+static void NoFunc (Widget w, XtPointer closure, XtPointer data)
 {
+    (void) w;
+    (void) closure;
+    (void) data;
     yn_answer = 0;
     yn_done = 1;
 }
@@ -294,10 +288,7 @@ DisplayDiscard (int type)
 void
 DisplayDeck (int numberLeft)
 {
-    char	buffer[512];
-
-    sprintf (buffer, "Cards: %3d", numberLeft);
-    displayString (deck_count, buffer);
+    Message(deck_count, "Cards: %d", numberLeft);
 }
 
 void
@@ -491,13 +482,15 @@ static void draw_action (Widget w, XEvent *e, String *p, Cardinal *n)
 
 static void play_action (Widget w, XEvent *e, String *p, Cardinal *n)
 {
-    (void) p;
-    (void) n;
+    (void) w;
+    (void) e;
     card_action (p, n, M_PLAY);
 }
 
 static void reasonable_action (Widget w, XEvent *e, String *p, Cardinal *n)
 {
+    (void) w;
+    (void) e;
     card_action (p, n, M_REASONABLE);
 }
 
@@ -515,18 +508,27 @@ static void order_action (Widget w, XEvent *e, String *p, Cardinal *n)
 static void quit_action (Widget w, XEvent *e, String *p, Cardinal *n)
 {
     (void) w;
+    (void) e;
+    (void) p;
+    (void) n;
     DoQuit (w, (XtPointer) 0, (XtPointer) 0);
 }
 
 static void save_action (Widget w, XEvent *e, String *p, Cardinal *n)
 {
     (void) w;
+    (void) e;
+    (void) p;
+    (void) n;
     DoSave (w, (XtPointer) 0, (XtPointer) 0);
 }
 
 static void restore_action (Widget w, XEvent *e, String *p, Cardinal *n)
 {
     (void) w;
+    (void) e;
+    (void) p;
+    (void) n;
     DoRestore (w, (XtPointer) 0, (XtPointer) 0);
 }
 
@@ -545,45 +547,67 @@ XtActionsRec actions[] = {
     { "milleRestore", restore_action, },
 };
 
-static void
-DisplayCallback (Widget w, XtPointer closure, XtPointer data)
-{
-    HandDisplayPtr  display = (HandDisplayPtr) data;
-    XRectangle	    *clip = 0;
-
-    if (display->clipped)
-	clip = &display->clip;
-    if ((intptr_t) display->private == -2)
-	drawIm (XtDisplay (w), XtWindow (w), &deck, display->x, display->y, clip);
-    else
-	displayCard (XtDisplay (w), XtWindow (w), (intptr_t) display->private,
-		     display->x, display->y, clip);
-}
-
 static void
 InputCallback (Widget w, XtPointer closure, XtPointer data)
 {
     HandInputPtr    input = (HandInputPtr) data;
-    String	    type;
+    String arg = "";
 
-    if (w == human_hand) {
-	Movetype = M_REASONABLE;
-	if (*input->num_params) {
-	    type = *input->params;
-	    if (!strcmp (type, "play"))
-		Movetype = M_PLAY;
-	    else if (!strcmp (type, "reasonable"))
-		Movetype = M_REASONABLE;
-	    else if (!strcmp (type, "discard"))
-		Movetype = M_DISCARD;
-	}
-	Card_no = input->col;
-	getmove_done = 1;
-    }
-    else if (w == deck_hand)
-    {
-	Movetype = M_DRAW;
-	getmove_done = 1;
+    (void) w;
+    (void) closure;
+    if (*input->num_params > 0)
+        arg = input->params[0];
+
+    switch (input->action) {
+    default:
+        break;
+    case HandActionClick:
+        animate_enable(1);
+        if (input->start.w == human_hand)
+        {
+            if (strcmp(arg, "discard") == 0)
+                Movetype = M_DISCARD;
+            else
+                Movetype = M_REASONABLE;
+            getmove_done = 1;
+            Card_no = input->start.col;
+        }
+        else if (input->start.w == deck_hand)
+        {
+            Movetype = M_DRAW;
+            getmove_done = 1;
+            Card_no = input->start.col;
+        }
+        break;
+    case HandActionDrag:
+        if (input->start.w == input->current.w)
+            break;
+        animate_enable(0);
+        if (input->start.w == human_hand) {
+            if (input->current.w == human_miles ||
+                input->current.w == human_play ||
+                input->current.w == human_safeties)
+            {
+                Movetype = M_PLAY;
+                getmove_done = 1;
+                Card_no = input->start.col;
+            }
+            else if (input->current.w == deck_hand)
+            {
+                Movetype = M_DISCARD;
+                getmove_done = 1;
+                Card_no = input->start.col;
+            }
+            break;
+        }
+        else if (input->start.w == deck_hand)
+        {
+            if (input->current.w == human_hand) {
+                Movetype = M_DRAW;
+                getmove_done = 1;
+                Card_no = input->start.col;
+            }
+        }
     }
 }
 
@@ -595,14 +619,17 @@ make_hand (char *name, Widget parent, int rows, int cols, Bool overlap_rows)
     Cardinal    i = 0;
     int		display_x, display_y;
 
-    XtSetArg (args[i], XtNcardWidth, WIDTH); i++;
-    XtSetArg (args[i], XtNcardHeight, HEIGHT); i++;
+    XtSetArg (args[i], XtNcardWidth, WIDTH * scale); i++;
+    XtSetArg (args[i], XtNcardHeight, HEIGHT * scale); i++;
     XtSetArg (args[i], XtNnumRows, rows); i++;
     XtSetArg (args[i], XtNnumCols, cols); i++;
+    XtSetArg (args[i], XtNinternalBorderWidth, WIDTH * scale / 20); i++;
     if (!overlap_rows) {
-	XtSetArg (args[i], XtNrowOffset, HEIGHT + WIDTH/10); i++;
+	XtSetArg (args[i], XtNrowOffset, HEIGHT * scale + WIDTH * scale/20); i++;
+    } else {
+	XtSetArg (args[i], XtNrowOffset, WIDTH * scale/20); i++;
     }
-    XtSetArg (args[i], XtNcolOffset, WIDTH + WIDTH/10); i++;
+    XtSetArg (args[i], XtNcolOffset, WIDTH * scale + WIDTH * scale/20); i++;
     display_x = 0;
     display_y = 0;
     if (rows == 1)
@@ -611,10 +638,10 @@ make_hand (char *name, Widget parent, int rows, int cols, Bool overlap_rows)
 	display_y = 8;
     XtSetArg (args[i], XtNdisplayX, display_x); i++;
     XtSetArg (args[i], XtNdisplayY, display_y); i++;
-    XtSetArg (args[i], XtNdisplayWidth, WIDTH - display_x * 2); i++;
-    XtSetArg (args[i], XtNdisplayHeight, HEIGHT - display_y * 2); i++;
-    hand = XtCreateManagedWidget (name, handWidgetClass, parent, args, i);
-    XtAddCallback (hand, XtNdisplayCallback, DisplayCallback, (XtPointer) hand);
+    XtSetArg (args[i], XtNdisplayWidth, WIDTH * scale - display_x * 2); i++;
+    XtSetArg (args[i], XtNdisplayHeight, HEIGHT * scale - display_y * 2); i++;
+    hand = XtCreateManagedWidget (name, milleCardsWidgetClass, parent, args, i);
+    XtAddCallback (hand, XtNinputCallback, InputCallback, NULL);
     return hand;
 }
 
@@ -625,6 +652,7 @@ DoRestore (Widget w, XtPointer closure, XtPointer data)
     (void) closure;
     (void) data;
     rest();
+    prboard();
 }
 
 static void
@@ -652,16 +680,18 @@ CreateMenu (Widget parent, char *name, struct menuEntry *entries, int count)
     Widget  entry;
     int	    i;
 
-    menu = XtCreatePopupShell (name, simpleMenuWidgetClass,
+    menu = XtCreatePopupShell (name, ksimpleMenuWidgetClass,
 			       parent, NULL, ZERO);
     for (i = 0; i < count; i++) {
 	entry = XtCreateManagedWidget (entries[i].name,
-				       smeBSBObjectClass, menu, NULL, ZERO);
+				       ksmeBSBObjectClass, menu, NULL, ZERO);
 	XtAddCallback (entry, XtNcallback, entries[i].function, NULL);
     }
     return menu;
 }
 
+int		clip_cards;
+
 void
 init_ui (int *argc, char **argv)
 {
@@ -670,17 +700,18 @@ init_ui (int *argc, char **argv)
     XGCValues	gcv;
     Colormap	def_cm;
     extern double	animation_speed;
-    extern int		clip_cards;
     unsigned long	gcmask;
-    Pixmap		grayStipple;
+    Pixmap		grayStipple = None;
     Arg			arg[2];
     Visual		*visual;
 
-#ifdef APPDEFAULTS
-    setenv("XAPPLRESDIR", APPDEFAULTS, 1);
-#endif
+    toplevel = XkwInitialize ("Mille", 0, 0,
+			      argc, argv, True, defaultResources);
+
+    Arg	args[1];
+    XtSetArg(args[0], XtNinput, True);
+    XtSetValues(toplevel, args, ONE);
 
-    toplevel = XtInitialize (argv[0], "Mille", 0, 0, argc, argv);
     dpy = XtDisplay (toplevel);
     screen = DefaultScreen(dpy);
     def_cm = DefaultColormap(dpy, screen);
@@ -756,51 +787,53 @@ init_ui (int *argc, char **argv)
 	init_mono_cards();
     layout = XtCreateManagedWidget ("layout", layoutWidgetClass, toplevel, NULL, (Cardinal) 0);
     menu_bar = XtCreateManagedWidget ("menuBar", boxWidgetClass, layout, NULL, ZERO);
-    file_menu_button = XtCreateManagedWidget ("fileMenuButton", menuButtonWidgetClass, menu_bar, NULL, ZERO);
+    file_menu_button = XtCreateManagedWidget ("fileMenuButton", kmenuButtonWidgetClass, menu_bar, NULL, ZERO);
     file_menu = CreateMenu (file_menu_button, "fileMenu", fileMenuEntries, XtNumber (fileMenuEntries));
     computer_play = make_hand ("computerPlay", layout, 3, 7, True);
     computer_miles = XtCreateManagedWidget ("computerMiles", thermoWidgetClass, layout, NULL, ZERO);
     computer_safeties = make_hand ("computerSafeties", layout, 2, 2, False);
-    computer_safety_label = XtCreateManagedWidget ("computerSafetyLabel", labelWidgetClass, layout, NULL, ZERO);
-    deck_count = XtCreateManagedWidget ("deckCount", labelWidgetClass, layout, NULL, ZERO);
+    computer_safety_label = XtCreateManagedWidget ("computerSafetyLabel", klabelWidgetClass, layout, NULL, ZERO);
+    deck_count = XtCreateManagedWidget ("deckCount", klabelWidgetClass, layout, NULL, ZERO);
     deck_hand = make_hand ("deck", layout, 1, 2, False);
     XtSetArg (arg[0], XtNnumRows, 13);
     XtSetArg (arg[1], XtNnumCols, 48);
     score = XtCreateManagedWidget ("score", padWidgetClass, layout,
 				   arg, TWO);
-    XtAddCallback (deck_hand, XtNinputCallback, InputCallback, NULL);
     HandAddCard (deck_hand, (XtPointer) -2, 0, 0, XkwHandDefaultOffset);
-    message = XtCreateManagedWidget ("message", labelWidgetClass, layout, NULL, ZERO);
-    errors = XtCreateManagedWidget ("errors", labelWidgetClass, layout, NULL, ZERO);
+    message = XtCreateManagedWidget ("message", klabelWidgetClass, layout, NULL, ZERO);
+    errors = XtCreateManagedWidget ("errors", klabelWidgetClass, layout, NULL, ZERO);
     human_miles = XtCreateManagedWidget ("humanMiles", thermoWidgetClass, layout, NULL, ZERO);
     human_play = make_hand ("humanPlay", layout, 3, 7, True);
     human_hand = make_hand ("humanHand", layout, 1, 7, False);
-    XtAddCallback (human_hand, XtNinputCallback, InputCallback, NULL);
     human_safeties = make_hand ("humanSafeties", layout, 2, 2, False);
-    human_safety_label = XtCreateManagedWidget ("humanSafetyLabel", labelWidgetClass, layout, NULL, ZERO);
+    human_safety_label = XtCreateManagedWidget ("humanSafetyLabel", klabelWidgetClass, layout, NULL, ZERO);
     newscore ();
     prscore (FALSE);
     XtRealizeWidget (toplevel);
+
+    XkwSetIcon(toplevel, icon.svg);
+
     yes_or_no_shell = XtCreatePopupShell ("yesOrNo", transientShellWidgetClass,
 			        toplevel, NULL, ZERO);
     yes_or_no_dialog = XtCreateManagedWidget ("yesOrNoDialog", layoutWidgetClass,
 				yes_or_no_shell, NULL, ZERO);
-    yes_or_no_label = XtCreateManagedWidget ("yesOrNoLabel", labelWidgetClass,
+    yes_or_no_label = XtCreateManagedWidget ("yesOrNoLabel", klabelWidgetClass,
 				yes_or_no_dialog, NULL, ZERO);
-    XawDialogAddButton (yes_or_no_dialog, "yesOrNoOk", YesFunc, NULL);
-    XawDialogAddButton (yes_or_no_dialog, "yesOrNoNo", NoFunc, NULL);
+    XkwDialogAddButton (yes_or_no_dialog, "yesOrNoOk", YesFunc, NULL);
+    XkwDialogAddButton (yes_or_no_dialog, "yesOrNoNo", NoFunc, NULL);
     XtRealizeWidget (yes_or_no_shell);
     prompted_shell = XtCreatePopupShell ("prompted", transientShellWidgetClass,
 				toplevel, NULL, ZERO);
     XtSetArg (arg[0], XtNvalue, "");
     prompted_dialog = XtCreateManagedWidget ("promptedDialog", layoutWidgetClass,
 				prompted_shell, arg, ONE);
-    prompted_label = XtCreateManagedWidget ("promptedLabel", labelWidgetClass,
+    prompted_label = XtCreateManagedWidget ("promptedLabel", klabelWidgetClass,
 				prompted_dialog, NULL, ZERO);
-    prompted_value = XtCreateManagedWidget ("promptedValue", asciiTextWidgetClass,
+    prompted_value = XtCreateManagedWidget ("promptedValue", ktextLineWidgetClass,
 				prompted_dialog, NULL, ZERO);
-    XawDialogAddButton (prompted_dialog, "promptedOk", YesFunc, NULL);
-    XawDialogAddButton (prompted_dialog, "promptedCancel", NoFunc, NULL);
+    XtAddCallback(prompted_value, XtNcallback, YesFunc, NULL);
+    XkwDialogAddButton (prompted_dialog, "promptedOk", YesFunc, NULL);
+    XkwDialogAddButton (prompted_dialog, "promptedCancel", NoFunc, NULL);
     XtRealizeWidget (prompted_shell);
 }
 
@@ -916,7 +949,7 @@ getmove(void)
 	XtNextEvent (&event);
 	XtDispatchEvent (&event);
     }
-    Message ("");
+    MilleMessage ("");
     Error ("", "");
 }
 
@@ -939,12 +972,11 @@ void
 prboard(void)
 {
 
-	register PLAY	*pp;
-	register int	i, k;
+	PLAY	*pp;
 
-	for (k = 0; k < 2; k++) {
+	for (int k = 0; k < 2; k++) {
 		pp = &Player[k];
-		for (i = 0; i < NUM_SAFE; i++)
+		for (int i = 0; i < NUM_SAFE; i++)
 			if (pp->safety[i] == S_PLAYED) {
 				if (k == 0) {
 					HumanSafety (i + S_CONV, i);
@@ -959,8 +991,8 @@ prboard(void)
 			ComputerBattle (pp->battle);
 			ComputerSpeed (pp->speed);
 		}
-		for (i = C_25; i <= C_200; i++) {
-			register int		end;
+		for (int i = C_25; i <= C_200; i++) {
+			int		end;
 
 			end = pp->nummiles[i];
 			if (k == 0)
@@ -971,7 +1003,7 @@ prboard(void)
 	}
 	prscore(TRUE);
 	pp = &Player[PLAYER];
-	for (i = 0; i < HAND_SZ; i++) {
+	for (int i = 0; i < HAND_SZ; i++) {
 		HumanHand (pp->hand[i], i);
 	}
 	DisplayDeck (Topcard - Deck);
@@ -988,6 +1020,9 @@ prboard(void)
 void
 stand(int y, int x, char *str)
 {
+    (void) y;
+    (void) x;
+    (void) str;
 }
 
 void
@@ -1003,11 +1038,11 @@ InScore (int line, int player, char *text)
 void
 prscore(bool for_real)
 {
-
-	register PLAY	*pp;
-	register char	*Score_fmt = "%4d  ";
+	PLAY		*pp;
+	const char Score_fmt[] = "%4d  ";
 	char		buffer[512];
 
+        (void) for_real;
 	ComputerDistance (Player[1].mileage);
 	HumanDistance (Player[0].mileage);
 
